#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import annotations

import asyncio
import re
from dataclasses import dataclass

from lvmieb.exceptions import LvmIebError


__all__ = ["PressureTransducer"]


@dataclass
class PressureTransducer:
    """Communicates with a SENS4 pressure transducer.

    Parameters
    ----------
    spec
        The spectrograph associated with this device.
    camera
        The camera (red, blue, NIR) that this device monitors.
    host
        The host that provides the TCP server.
    port
        The port on which the TCP server is running.
    device_id
        The ID of the device.
    disabled
        Whether this device is disabled.

    """

    spec: str
    camera: str
    host: str
    port: int
    disabled: bool = False
    device_id: int = 254

    TIMEOUT: float = 3

    async def _read(self, query_string: str = "P"):
        """Queries the transducer."""

        try:
            r, w = await asyncio.wait_for(
                asyncio.open_connection(self.host, self.port),
                self.TIMEOUT,
            )
        except Exception as err:
            raise LvmIebError(
                f"Transducer {self.camera}: failed connecting to device: {err}"
            )

        command = "@" + str(self.device_id) + query_string + "?\\"
        w.write(command.encode())
        await w.drain()

        try:
            reply = await asyncio.wait_for(r.readuntil(b"\\"), self.TIMEOUT)
            match = re.search(r"@[0-9]{1,3}ACK([0-9.E+-]+)\\$".encode(), reply)
            if not match:
                raise ValueError("Cannot parse reply.")
            return float(match.groups()[0])
        finally:
            w.close()
            await w.wait_closed()

    async def read_pressure(self):
        """Reads the pressure from the transducer."""

        return await self._read("P")

    async def read_temperature(self):
        """Reads the temperature from the transducer."""

        return await self._read("T")



from __future__ import annotations

from typing import TYPE_CHECKING

import click
import numpy

from . import parser


if TYPE_CHECKING:
    from lvmieb.actor.actor import ControllersType, IEBCommand
    from lvmieb.controller.controller import IEBController


async def read_transducer(
    controller: IEBController,
    camera: str,
    measurement: str = "pressure",
):
    """Reads a pressure transducer value."""

    NRETRIES = 3

    pressure_transducer = controller.pressure[camera]

    if measurement == "pressure":
        func = pressure_transducer.read_pressure
    elif measurement == "temperature":
        func = pressure_transducer.read_temperature
    else:
        raise ValueError(f"Unknown measurement {measurement}")

    for iretry in range(NRETRIES):
        try:
            result = await func()
            return result
        except Exception as err:
            if iretry < NRETRIES - 1:
                continue
            else:
                raise err


@parser.group()
def transducer(*args):
    """Reports pressure transducer values.."""
    pass


@transducer.command()
@click.argument("spectro", type=str, required=False)
async def status(
    command: IEBCommand,
    controllers: ControllersType,
    spectro: str | None = None,
):
    """Returns the status of transducer."""

    pres_result = {}

    for controller_name in controllers:
        if spectro is not None and controller_name != spectro:
            continue

        controller = controllers[controller_name]

        # We read cameras and measurements sequentially instead of with a gather
        # to avoid too many concurrent accesses to the hardware. This could not matter
        # that much anymore, though.
        for cam in controller.pressure:
            pressure_transducer = controller.pressure[cam]

            for measurement in ["pressure", "temperature"]:
                try:
                    if pressure_transducer.disabled:
                        value = numpy.nan
                    else:
                        value = await read_transducer(controller, cam, measurement)
                except Exception as err:
                    command.warning(f"Failed to read {measurement} from {cam}: {err}")
                    value = numpy.nan

                pres_result[f"{cam}_{measurement}"] = value

    command.finish(transducer=pres_result)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Changgon Kim, Taeeun Kim, Mingyeong YANG (mingyeong@khu.ac.kr)
# @Date: 2021-03-22
# @Filename: actor.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import os
import pathlib
import warnings
from copy import deepcopy

from typing import ClassVar

from clu import Command
from clu.actor import AMQPActor
from sdsstools.configuration import read_yaml_file

from lvmieb import __version__, config
from lvmieb.controller.controller import IEBController
from lvmieb.controller.depth import DepthGauges
from lvmieb.exceptions import LvmIebUserWarning

from .commands import parser as lvm_command_parser


__all__ = ["IEBActor", "IEBCommand", "ControllersType"]


class IEBActor(AMQPActor):
    """LVM Spectrograph controller actor.

    In addition to the normal arguments and keyword parameters for
    `~clu.actor.AMQPActor`, the class accepts the following parameters.

    Parameters
    ----------
    controllers
        The list of '.IEBController' instances to manage.

    """

    parser = lvm_command_parser
    BASE_CONFIG: ClassVar[str | dict | None] = config

    def __init__(
        self,
        *args,
        controllers: tuple[IEBController, ...] = (),
        depth_gauges: DepthGauges | None = None,
        **kwargs,
    ):
        self.controllers = {c.spec: c for c in controllers}
        self.depth_gauges = depth_gauges

        self.version = __version__

        super().__init__(*args, **kwargs)

    async def start(self, **kwargs):  # pragma: no cover
        """Starts the actor connection to RabbitMQ."""

        # We wait to set this until here to make sure all the controllers
        # have been created.
        self.parser_args = [self.controllers]

        return await super().start(**kwargs)

    @classmethod
    def from_config(cls, config: dict | str | None, *args, **kwargs):
        """Creates an actor from a configuration file."""

        if config is None:
            if cls.BASE_CONFIG is None:
                raise RuntimeError("The class does not have a base configuration.")
            config = cls.BASE_CONFIG

        if isinstance(config, (str, pathlib.Path)):
            config = dict(read_yaml_file(str(config)))

        if "schema" in config["actor"]:
            schema = config["actor"]["schema"]
            if not os.path.isabs(schema):
                schema = os.path.join(
                    os.path.abspath(os.path.join(os.path.dirname(__file__), "../")),
                    schema,
                )
            config["actor"]["schema"] = schema

        instance = super().from_config(config["actor"], *args, **kwargs)

        controllers: list[IEBController] = []

        for spec in config.get("enabled_specs", []):
            if "specs" not in config or spec not in config["specs"]:
                warnings.warn(
                    f"Cannot find configuration for {spec!r}.",
                    LvmIebUserWarning,
                )
                continue

            spec_config = config["specs"][spec].copy()
            wago_modules = deepcopy(config.get("wago_modules", {}))

            controller = IEBController.from_config(
                spec,
                spec_config,
                wago_modules=wago_modules,
            )
            controllers.append(controller)

        instance.controllers = {contr.spec: contr for contr in controllers}

        if (depth_gauges := config.get("depth_gauges", None)) is not None:
            instance.depth_gauges = DepthGauges(**depth_gauges.copy())

        return instance


IEBCommand = Command[IEBActor]
ControllersType = dict[str, IEBController]
